@@ -22,18 +22,18 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
             connection.client_id()
         ),
     );
-    connection.set_voice_rate(100);
+    connection.set_voice_rate(100)?;
     connection.say(Priority::Important, "This is faster.");
-    connection.set_voice_rate(0);
-    connection.set_spelling(true);
+    connection.set_voice_rate(0)?;
+    connection.set_spelling(true)?;
     connection.say(Priority::Important, "This is spelled.");
-    connection.set_spelling(false);
-    connection.set_punctuation(Punctuation::All);
+    connection.set_spelling(false)?;
+    connection.set_punctuation(Punctuation::All)?;
     connection.say(
         Priority::Important,
         "This statement, unlike others, has punctuation that is spoken!",
     );
-    connection.set_punctuation(Punctuation::None);
+    connection.set_punctuation(Punctuation::None)?;
     let mut _input = String::new();
     io::stdin().read_line(&mut _input).unwrap();
     Ok(())
